/******************************************************************************
                                pnmquant.c
*******************************************************************************
  Quantize the colors in an image -- the output image has fewer colors than
  the input.

  By Bryan Henderson, San Jose CA, November 2001

  Derived from Ppmquant, which has the following copyright notice and 
  license:

  Copyright (C) 1989, 1991 by Jef Poskanzer.
  
  Permission to use, copy, modify, and distribute this software and its
  documentation for any purpose and without fee is hereby granted, provided
  that the above copyright notice appear in all copies and that both that
  copyright notice and this permission notice appear in supporting
  documentation.  This software is provided "as is" without express or
  implied warranty.

******************************************************************************/

#include <limits.h>

#include "pam.h"
#include "pammap.h"

#define MAXCOLORS 32767u
#define MAXDEPTH 3

enum methodForLargest {LARGE_NORM, LARGE_LUM};

enum methodForRep {REP_CENTER_BOX, REP_AVERAGE_COLORS, REP_AVERAGE_PIXELS};

#define FS_SCALE 1024

typedef struct box* box_vector;
struct box {
    int ind;
    int colors;
    int sum;
};

struct fserr {
    long** thiserr;
    long** nexterr;
    bool fsForward;
};


struct cmdlineInfo {
    /* All the information the user supplied in the command line,
       in a form easy for the program to use.
    */
    char *inputFilespec;  /* Filespec of input file */
    char *mapFilespec;    /* Filespec of colormap file; NULL if none */
    unsigned int newcolors;   
        /* Number of colors argument; meaningless if mapFilespec non-null */
    unsigned int floyd;   /* Boolean: -floyd/-fs option */
    enum methodForLargest methodForLargest; 
        /* -spreadintensity/-spreadluminosity options */
    enum methodForRep methodForRep;
        /* -center/-meancolor/-meanpixel options */
};


static int __inline__ sqr(const int x) {
    return x*x;
}



static void
parseCommandLine(int argc, char ** argv,
                 struct cmdlineInfo *cmdlineP) {
/*----------------------------------------------------------------------------
   Note that the file spec array we return is stored in the storage that
   was passed to us as the argv array.
-----------------------------------------------------------------------------*/
    optStruct3 opt;  /* set by OPTENT3 */
    optEntry *option_def = malloc(100*sizeof(optEntry));
    unsigned int option_def_index;

    unsigned int mapfileSpec;
    unsigned int nofloyd, spreadbrightness, spreadluminosity;
    unsigned int center, meancolor, meanpixel;
    
    option_def_index = 0;   /* incremented by OPTENT3 */
    OPTENT3(0,   "floyd",       OPT_FLAG,   
            NULL,                       &cmdlineP->floyd, 0);
    OPTENT3(0,   "fs",          OPT_FLAG,   
            NULL,                       &cmdlineP->floyd, 0);
    OPTENT3(0,   "nofloyd",     OPT_FLAG,   
            NULL,                       &nofloyd, 0);
    OPTENT3(0,   "nofs",        OPT_FLAG,   
            NULL,                       &nofloyd, 0);
    OPTENT3(0,   "mapfile",     OPT_STRING, 
            &cmdlineP->mapFilespec,     &mapfileSpec, 0);
    OPTENT3(0,   "spreadbrightness", OPT_FLAG,   
            NULL,                       &spreadbrightness, 0);
    OPTENT3(0,   "spreadluminosity", OPT_FLAG,   
            NULL,                       &spreadluminosity, 0);
    OPTENT3(0,   "center",           OPT_FLAG,   
            NULL,                       &center,           0);
    OPTENT3(0,   "meancolor",        OPT_FLAG,   
            NULL,                       &meancolor,        0);
    OPTENT3(0,   "meanpixel",        OPT_FLAG,   
            NULL,                       &meanpixel,        0);

    opt.opt_table = option_def;
    opt.short_allowed = FALSE;  /* We have no short (old-fashioned) options */
    opt.allowNegNum = FALSE;  /* We may have parms that are negative numbers */

    pm_optParseOptions3(&argc, argv, opt, sizeof(opt), 0);
        /* Uses and sets argc, argv, and some of *cmdlineP and others. */

    if (cmdlineP->floyd && nofloyd)
        pm_error("You cannot specify both -floyd and -nofloyd options.");

    if (spreadbrightness && spreadluminosity) 
        pm_error("You cannot specify both -spreadbrightness and "
                 "spreadluminosity.");
    if (spreadluminosity)
        cmdlineP->methodForLargest = LARGE_LUM;
    else
        cmdlineP->methodForLargest = LARGE_NORM;

    if (center + meancolor + meanpixel > 1)
        pm_error("You can specify only one of -center, -meancolor, and "
                 "-meanpixel.");
    if (meancolor)
        cmdlineP->methodForRep = REP_AVERAGE_COLORS;
    else if (meanpixel) 
        cmdlineP->methodForRep = REP_AVERAGE_PIXELS;
    else
        cmdlineP->methodForRep = REP_CENTER_BOX;

    if (!mapfileSpec)
        cmdlineP->mapFilespec = NULL;

    if (mapfileSpec) {
        if (argc-1 == 0)
            cmdlineP->inputFilespec = "-";
        else if (argc-1 != 1) 
            pm_error("When you specify the -mapfile option, the only "
                     "argument allowed is the input file specification.  "
                     "You specified %d arguments.", argc-1);
        else
            cmdlineP->inputFilespec = argv[1];
    } else {
        if (argc-1 > 2)
            pm_error("Program takes at most two arguments: number of colors "
                     "and input file specification.  "
                     "You specified %d arguments.", argc-1);
        else {
            if (argc-1 < 2)
                cmdlineP->inputFilespec = "-";
            else
                cmdlineP->inputFilespec = argv[2];

            if (argc-1 < 1)
                pm_error("You must specify the number of colors in the "
                         "output as an argument, or specify the -mapfile "
                         "option.");
            else {
                char * tail;
                long int const newcolors = strtol(argv[1], &tail, 10);
                if (*tail != '\0')
                    pm_error("The number of colors argument '%s' is not "
                             "a number.", argv[1]);
                else if (newcolors < 1)
                    pm_error("The number of colors must be positive");
                else if (newcolors == 1)
                    pm_error("The number of colors must be greater than 1.");
                else
                    cmdlineP->newcolors = newcolors;
            }
        }
    }
}



static int __inline__
compareplane(struct tupleint ** const comparandP, 
             struct tupleint ** const comparatorP,
             unsigned int       const plane) {
    return (*comparandP)->tuple[plane] - (*comparatorP)->tuple[plane];
}

static int
compareplane0(const void * const comparandP, const void * const comparatorP) {
    return compareplane((struct tupleint **)comparandP,
                        (struct tupleint **)comparatorP,
                        0);
}

static int
compareplane1(const void * const comparandP, const void * const comparatorP) {
    return compareplane((struct tupleint **)comparandP,
                        (struct tupleint **)comparatorP,
                        1);
}

static int
compareplane2(const void * const comparandP, const void * const comparatorP) {
    return compareplane((struct tupleint **)comparandP,
                        (struct tupleint **)comparatorP,
                        2);
}



typedef int qsort_comparison_fn(const void *, const void *);

/* This ought to be an array of functions, but the compiler complains at
   that for some reason.
*/
static struct {qsort_comparison_fn * fn;} samplecompare[MAXDEPTH];

static void
initSampleCompare(void) {
    samplecompare[0].fn = compareplane0;
    samplecompare[1].fn = compareplane1;
    samplecompare[2].fn = compareplane2;
}



static int
sumcompare(const void * const b1, const void * const b2) {
    return(((box_vector)b2)->sum - ((box_vector)b1)->sum);
}




/*
** Here is the fun part, the median-cut colormap generator.  This is based
** on Paul Heckbert's paper "Color Image Quantization for Frame Buffer
** Display", SIGGRAPH '82 Proceedings, page 297.
*/

static tupletable
newColorMap(const struct pam * const pamP, 
            unsigned int       const newcolors) {

    tupletable colormap;
    unsigned int i;

    colormap = pnm_alloctupletable(pamP, newcolors);

    for (i = 0; i < newcolors; ++i) {
        unsigned int plane;
        for (plane = 0; plane < pamP->depth; ++plane) 
            colormap[i]->tuple[plane] = 0;
    }
    return colormap;
}



static box_vector
newBoxVector(int const colors, int const sum, int const newcolors) {

    box_vector bv;
    bv = (box_vector) malloc(sizeof(struct box) * newcolors);
    if (bv == NULL)
        pm_error("out of memory allocating box vector table");

    /* Set up the initial box. */
    bv[0].ind = 0;
    bv[0].colors = colors;
    bv[0].sum = sum;

    return bv;
}



static void
findBoxBoundaries(struct pam * const pamP, 
                  tupletable   const colorfreqtable,
                  unsigned int const indx,
                  unsigned int const clrs,
                  sample             minval[],
                  sample             maxval[]) {
/*----------------------------------------------------------------------------
  Go through the box finding the minimum and maximum of each
  component - the boundaries of the box.
-----------------------------------------------------------------------------*/
    unsigned int plane;
    unsigned int i;

    for (plane = 0; plane < pamP->depth; ++plane) 
        minval[plane] = maxval[plane] = colorfreqtable[indx]->tuple[plane];
    
    for (i = 1; i < clrs; ++i) {
        unsigned int plane;
        for (plane = 0; plane < pamP->depth; ++plane) {
            sample const v = colorfreqtable[indx + i]->tuple[plane];
            if (v < minval[plane]) minval[plane] = v;
            if (v > maxval[plane]) maxval[plane] = v;
        } 
    }
}



static unsigned int
largestByNorm(struct pam * const pamP, sample minval[], sample maxval[]) {
    
    unsigned int largestDimension;
    unsigned int plane;

    sample largestSpreadSoFar = 0;  
    largestDimension = 0;
    for (plane = 0; plane < pamP->depth; ++plane) {
        sample const spread = maxval[plane]-minval[plane];
        if (spread > largestSpreadSoFar) {
            largestDimension = plane;
            largestSpreadSoFar = spread;
        }
    }
    return largestDimension;
}



static unsigned int 
largestByLuminosity(struct pam * const pamP, sample minval[], sample maxval[]) {
/*----------------------------------------------------------------------------
   This subroutine presumes that the tuple type is either 
   BLACKANDWHITE, GRAYSCALE, or RGB (which implies pamP->depth is 1 or 3).
   To save time, we don't actually check it.
-----------------------------------------------------------------------------*/
    unsigned int retval;

    if (pamP->depth == 1)
        retval = 0;
    else {
        /* An RGB tuple */
        unsigned int largestDimension;
        unsigned int plane;
        double largestSpreadSoFar;

        largestSpreadSoFar = 0.0;

        for (plane = 0; plane < 3; ++plane) {
            double const spread = 
                pnm_lumin_factor[plane] * (maxval[plane]-minval[plane]);
            if (spread > largestSpreadSoFar) {
                largestDimension = plane;
                largestSpreadSoFar = spread;
            }
        }
        retval = largestDimension;
    }
    return retval;
}



static void
centerBox(struct pam * const pamP, 
          int          const indx,
          int          const clrs,
          tupletable   const colorfreqtable, 
          tuple        const newTuple) {

    unsigned int plane;

    for (plane = 0; plane < pamP->depth; ++plane) {
        int minval, maxval;
        unsigned int i;

        minval = maxval = colorfreqtable[indx]->tuple[plane];
        
        for (i = 1; i < clrs; ++i) {
            int const v = colorfreqtable[indx + i]->tuple[plane];
            minval = min( minval, v);
            maxval = max( maxval, v);
        }
        newTuple[plane] = (minval + maxval) / 2;
    }
}



static void
averageColors(struct pam * const pamP, 
              int          const indx,
              int          const clrs,
              tupletable   const colorfreqtable, 
              tuple        const newTuple) {

    unsigned int plane;

    for (plane = 0; plane < pamP->depth; ++plane) {
        sample sum;
        int i;

        sum = 0;

        for (i = 0; i < clrs; ++i) 
            sum += colorfreqtable[indx+i]->tuple[plane];

        newTuple[plane] = sum / clrs;
    }
}



static void
averagePixels(struct pam * const pamP, 
              int          const indx,
              int          const clrs,
              tupletable   const colorfreqtable, 
              tuple        const newTuple) {

    unsigned int n;
        /* Number of tuples represented by the box */
    unsigned int plane;
    unsigned int i;

    /* Count the tuples in question */
    n = 0;  /* initial value */
    for (i = 0; i < clrs; ++i)
        n += colorfreqtable[indx + i]->value;


    for (plane = 0; plane < pamP->depth; ++plane) {
        sample sum;
        int i;

        sum = 0;

        for (i = 0; i < clrs; ++i) 
            sum += colorfreqtable[indx+i]->tuple[plane]
                * colorfreqtable[indx+i]->value;

        newTuple[plane] = sum / n;
    }
}



static tupletable
colormapFromBv(struct pam *      const pamP, 
               unsigned int      const newcolors,
               box_vector        const bv, 
               unsigned int      const boxes,
               tupletable        const colorfreqtable, 
               enum methodForRep const methodForRep) {
    /*
    ** Ok, we've got enough boxes.  Now choose a representative color for
    ** each box.  There are a number of possible ways to make this choice.
    ** One would be to choose the center of the box; this ignores any structure
    ** within the boxes.  Another method would be to average all the colors in
    ** the box - this is the method specified in Heckbert's paper.  A third
    ** method is to average all the pixels in the box. 
    */
    tupletable colormap;
    unsigned int bi;

    colormap = newColorMap(pamP, newcolors);

    for (bi = 0; bi < boxes; ++bi) {
        switch (methodForRep) {
        case REP_CENTER_BOX: 
            centerBox(pamP, bv[bi].ind, bv[bi].colors, colorfreqtable, 
                      colormap[bi]->tuple);
            break;
        case REP_AVERAGE_COLORS:
            averageColors(pamP, bv[bi].ind, bv[bi].colors, colorfreqtable,
                          colormap[bi]->tuple);
            break;
        case REP_AVERAGE_PIXELS:
            averagePixels(pamP, bv[bi].ind, bv[bi].colors, colorfreqtable,
                          colormap[bi]->tuple);
            break;
        default:
            pm_error("Internal error: invalid value of methodForRep: %d",
                     methodForRep);
        }
    }
    return colormap;
}



static tupletable
mediancut(struct pam * const pamP,
          tupletable   const colorfreqtable, 
          int          const colors, 
          int          const newcolors,
          enum methodForLargest const methodForLargest,
          enum methodForRep     const methodForRep) {
/*----------------------------------------------------------------------------
   If *pamP describes an image with 'colors' distinct tuple values in it,
   and colorfreqtable[] gives how many times each of those tuple values
   occurs, compute a set of only 'newcolors' colors that are close to the
   colors in colorfreqtable[].

   As a side effect, sort 'colorfreqtable'.
-----------------------------------------------------------------------------*/

    tupletable colormap;
    box_vector bv;
    int bi;
    int boxes;

    bv = newBoxVector(colors, pamP->width*pamP->height, newcolors);
    boxes = 1;


    /* Main loop: split boxes until we have enough. */
    while (boxes < newcolors) {
        unsigned int indx;
        unsigned int clrs;
        unsigned int sm;
        sample minval[MAXDEPTH];
        sample maxval[MAXDEPTH];
        unsigned int largestDimension;
            /* number of the plane with the largest spread */
        unsigned int medianIndex;
        int lowersum;
            /* Number of pixels whose value is "less than" the median */

        /* Find the first splittable box. */
        for (bi = 0; bi < boxes && bv[bi].colors < 2; ++bi);
        if (bi >= boxes)
            break;  /* ran out of colors! */
        indx = bv[bi].ind;
        clrs = bv[bi].colors;
        sm = bv[bi].sum;

        findBoxBoundaries(pamP, colorfreqtable, indx, clrs, minval, maxval);

        /*
        ** Find the largest dimension, and sort by that component.  I have
        ** included two methods for determining the "largest" dimension;
        ** first by simply comparing the range in RGB space, and second
        ** by transforming into luminosities before the comparison. 
        */
        switch (methodForLargest) {
        case LARGE_NORM: 
            largestDimension = largestByNorm(pamP, minval, maxval);
            break;
        case LARGE_LUM: 
            largestDimension = largestByLuminosity(pamP, minval, maxval);
            break;
        }
                                                    
        /* TODO: I think this sort should go after creating a box, not before
           splitting.  Because you need the sort to use the REP_CENTER_BOX
           method of choosing a color to represent the final boxes 
        */
        qsort((char*) &colorfreqtable[indx], clrs, 
              sizeof(colorfreqtable[indx]), 
              samplecompare[largestDimension].fn);

        {
            /* Now find the median based on the counts, so that about half the
               pixels (not colors, pixels) are in each subdivision.
            */
            unsigned int i;

            lowersum = colorfreqtable[indx]->value;  /* initial value */
            for (i = 1; i < clrs - 1 && lowersum < sm/2; ++i) {
                lowersum += colorfreqtable[indx + i]->value;
            }
            medianIndex = i;
        }
        /* Split the box, and sort to bring the biggest boxes to the top. */
        bv[bi].colors = medianIndex;
        bv[bi].sum = lowersum;
        bv[boxes].ind = indx + medianIndex;
        bv[boxes].colors = clrs - medianIndex;
        bv[boxes].sum = sm - lowersum;
        ++boxes;
        qsort((char*) bv, boxes, sizeof(struct box), sumcompare);
    }

    colormap = colormapFromBv(pamP, newcolors, bv, boxes, colorfreqtable, 
                              methodForRep);

    free(bv);
    return colormap;
}




static void
computeHistogram(struct pam *   const pamP, 
                 tuple **       const tuples,
                 tupletable *   const colorfreqtableP, 
                 unsigned int * const colorsP) {
/*----------------------------------------------------------------------------
  Attempt to make a histogram of the colors, unclustered.  If at first
  we don't succeed, lower maxval to increase color coherence and try
  again.  This will eventually terminate, with maxval at worst 15,
  since 32^3 is approximately MAXCOLORS.

  We actually change the maxval of the input 'tuples' and update the input
  *pamP to reflect the new maxval.
-----------------------------------------------------------------------------*/
    tupletable colorfreqtable;
    unsigned int colors;

    colorfreqtable = NULL;  /* initial value */
    
    while (colorfreqtable == NULL)
    {
        pm_message( "making histogram..." );
        colorfreqtable = pnm_computetuplefreqtable(pamP, tuples, 
                                                   MAXCOLORS, &colors);
        if (colorfreqtable == NULL) {
            sample const newmaxval = pamP->maxval / 2;
            int row;

            pm_message("too many colors!");
            pm_message("scaling colors from maxval=%d to maxval=%d "
                       "to improve clustering...",
                       pamP->maxval, newmaxval );
            for (row = 0; row < pamP->height; ++row) {
                int col;
                for (col = 0; col < pamP->width; ++col)
                    pnm_scaletuple(pamP, tuples[row][col], tuples[row][col], 
                                   newmaxval);
            }
            pamP->maxval = newmaxval;
        }
    }
    pm_message( "%d colors found", colors );
    *colorfreqtableP = colorfreqtable;
    *colorsP = colors;
}




static void
computeColorMapFromInput(struct pam * const pamP, tuple ** const tuples,
                         int const reqColors, 
                         enum methodForLargest const methodForLargest,
                         enum methodForLargest const methodForRep,
                         tupletable * const colormapP,
                         int * const colorsP) {
/*----------------------------------------------------------------------------
   Produce a colormap containing the colors that we will use in the output.
   Figure it out using the median cut technique, based on the colors (and
   their frequencies) in the input image described by *pamP and tuples[][].

   The output map will have 'newcolors' or fewer colors in it.

   Modify the maxval of 'tuples', and update *pamP to reflect it, if 
   necessary.
-----------------------------------------------------------------------------*/
    tupletable colorfreqtable;
    unsigned int colors;

    computeHistogram(pamP, tuples, &colorfreqtable, &colors);
       /* May modify tuples[][]/pamP to reduce the maxval */
    
    if (colors <= reqColors) {
        pm_message("Image already has few enough colors (<=%d).  "
                   "Keeping same colors.", reqColors);
        *colormapP = colorfreqtable;
        *colorsP = colors;
    } else {
        pm_message("choosing %d colors...", reqColors);
        *colormapP = mediancut(pamP, colorfreqtable, colors, reqColors,
                               methodForLargest, methodForRep);
        *colorsP = reqColors;
        pnm_freetupletable(pamP, colorfreqtable);
    }
}



static void
adjustDepth(struct pam * const pamP,
            tuple ***    const tupleArrayP,
            unsigned int const newDepth) {
/*----------------------------------------------------------------------------
   Change the depth of the image described by 'pamP' and
   (*tupleArray)[][] to newDepth to match a user-supplied color map.

   Update *pamP to reflect the change.

   If necessary, allocate a new tuple array, free the old one, and
   update *tupleArray to point to the new one.

   The only depth change we allow is from an RGB image of depth 3 to
   depth 1.  In that case, we change it to grayscale or black and white.

   For any other depth change request, we issue an error message and abort
   the program.  (It is technically possible to go from grayscale to RGB,
   but there seems to be little requirement for that, so we save the 
   programming effort).
-----------------------------------------------------------------------------*/
    if (newDepth != pamP->depth) {
        struct pam oldpam = *pamP;

        if (!stripeq(pamP->tuple_type, "RGB"))
            pm_error("Map file depth of %d is incompatible with image "
                     "tuple type of '%*s'.  Only type RGB images can have "
                     "their depths changed.", newDepth, 
                     sizeof(pamP->tuple_type), pamP->tuple_type);
        else if (newDepth != 1)
            pm_error("Map file depth is %d, while image depth is %d.  "
                     "The only depth change allowed is from 3 to 1.",
                     newDepth, pamP->depth);
        else {
            tuple ** const tupleArray = *tupleArrayP;
            tuple ** newTupleArray;
            int row;

            if (pamP->maxval == 1) {
                pm_message("Converting input fom RGB to black and white to "
                           "match colormap file.");
                pamP->format = RPBM_FORMAT;
                strcpy(pamP->tuple_type, "BLACKANDWHITE");
            } else {
                pm_message("Converting input from RGB to grayscale to match "
                           "colormap file");
                pamP->format = RPGM_FORMAT;
                strcpy(pamP->tuple_type, "GRAYSCALE");
            }
            pamP->depth = newDepth;

            newTupleArray = pnm_allocpamarray(pamP);

            for (row = 0; row < pamP->height; ++row) {
                int col;
                for (col = 0; col < pamP->width; ++col) {
                    unsigned int plane;
                    double grayvalue;
                    grayvalue = 0.0;  /* initial value */
                    for (plane = 0; plane < oldpam.depth; ++plane)
                        grayvalue += 
                            pnm_lumin_factor[plane] * 
                            tupleArray[row][col][plane];
                    newTupleArray[row][col][0] = (int) (grayvalue + 0.5);
                }
            }
            pnm_freepamarray(*tupleArrayP, &oldpam);
            *tupleArrayP = newTupleArray;
        }
    }
}




static void
adjustMaxval(struct pam * const pamP,
             tuple **     const tupleArray,
             unsigned int const newMaxval) {
/*----------------------------------------------------------------------------
   Change the maxval of the image described by 'pamP' and
   (*tupleArray)[][] to newMaxval to match a user-supplied color map.

   Update *pamP to reflect the change.

-----------------------------------------------------------------------------*/
    if (newMaxval != pamP->maxval) {
        struct pam const oldpam = *pamP;
        int row;

        pm_message("rescaling input image to match colormap maxval (%d->%d)",
                   oldpam.maxval, newMaxval);

        pamP->maxval = newMaxval;

        for (row = 0; row < pamP->height; ++row) {
            int col;
            for (col = 0; col < pamP->width; ++col)
                pnm_scaletuple(&oldpam, 
                               tupleArray[row][col], tupleArray[row][col], 
                               pamP->maxval);
        }
    }
}



static void
computeColorMapFromMap(struct pam * const mappamP, 
                       tuple **     const maptuples, 
                       tupletable * const colormapP,
                       int *        const newcolorsP) {
/*----------------------------------------------------------------------------
   Produce a colormap containing the colors that we will use in the output.

   Make it include exactly those colors that are in the image
   described by *mappamP and maptuples[][].

   Return the number of colors in the returned colormap as *newcolorsP.
-----------------------------------------------------------------------------*/
    unsigned int colors; 

    if (mappamP->width == 0 || mappamP->height == 0)
        pm_error("colormap file contains no pixels");

    *colormapP = 
        pnm_computetuplefreqtable(mappamP, maptuples, MAXCOLORS, &colors);
    if (*colormapP == NULL)
        pm_error("too many colors in colormap!");
    pm_message("%d colors found in colormap", colors);
    *newcolorsP = colors;
}



static void
initFserr(struct pam * const pamP, struct fserr * const fserrP) {
/*----------------------------------------------------------------------------
   Initialize the Floyd-Steinberg error vectors
-----------------------------------------------------------------------------*/
    unsigned int plane;

    unsigned int const fserrSize = pamP->width + 2;

    fserrP->thiserr = malloc(pamP->depth * sizeof(long*));
    if (fserrP->thiserr == NULL)
        pm_error("Out of memory allocating Floyd-Steinberg structures");
    fserrP->nexterr = malloc(pamP->depth * sizeof(long*));
    if (fserrP->nexterr == NULL)
        pm_error("Out of memory allocating Floyd-Steinberg structures");
    
    for (plane = 0; plane < pamP->depth; ++plane) {
        fserrP->thiserr[plane] = (long*) malloc(fserrSize * sizeof(long));
        if (fserrP->thiserr[plane] == NULL)
            pm_error("Out of memory allocating Floyd-Steinberg structures");
        fserrP->nexterr[plane] = (long*) malloc(fserrSize * sizeof(long));
        if (fserrP->nexterr[plane] == NULL)
            pm_error("Out of memory allocating Floyd-Steinberg structures");
    }

    srand((int)(time(0) ^ getpid()));

    {
        int col;

        for (col = 0; col < fserrSize; ++col) {
            unsigned int plane;
            for (plane = 0; plane < pamP->depth; ++plane) 
                fserrP->thiserr[plane][col] = 
                    rand() % (FS_SCALE * 2) - FS_SCALE;
                    /* (random errors in [-1 .. 1]) */
        }
    }
    fserrP->fsForward = TRUE;
}



static void
floydInitRow(struct pam * const pamP, struct fserr * const fserrP) {

    int col;
    
    for (col = 0; col < pamP->width + 2; ++col) {
        unsigned int plane;
        for (plane = 0; plane < pamP->depth; ++plane) 
            fserrP->nexterr[plane][col] = 0;
    }
}



static void
floydAdjustColor(struct pam *   const pamP, 
                 tuple          const tuple, 
                 struct fserr * const fserrP, 
                 int            const col) {
/*----------------------------------------------------------------------------
  Use Floyd-Steinberg errors to adjust actual color.
-----------------------------------------------------------------------------*/
    unsigned int plane;

    for (plane = 0; plane < pamP->depth; ++plane) {
        sample const s =
            tuple[plane] + fserrP->thiserr[plane][col+1] / FS_SCALE;
        tuple[plane] = max(0, min(pamP->maxval, s));
    }
}



static void
floydPropagateErr(struct pam *   const pamP, 
                  struct fserr * const fserrP, 
                  int            const col, 
                  tuple          const oldtuple, 
                  tuple          const newtuple) {
/*----------------------------------------------------------------------------
  Propagate Floyd-Steinberg error terms.
-----------------------------------------------------------------------------*/

    if (fserrP->fsForward) {
        unsigned int plane;
        for (plane = 0; plane < pamP->depth; ++plane) {
            long const err = (oldtuple[plane] - newtuple[plane]) * FS_SCALE;
            fserrP->thiserr[plane][col + 2] += ( err * 7 ) / 16;
            fserrP->nexterr[plane][col    ] += ( err * 3 ) / 16;
            fserrP->nexterr[plane][col + 1] += ( err * 5 ) / 16;
            fserrP->nexterr[plane][col + 2] += ( err     ) / 16;
        }
    } else {
        unsigned int plane;
        for (plane = 0; plane < pamP->depth; ++plane) {
            long const err = (oldtuple[plane] - newtuple[plane]) * FS_SCALE;
            fserrP->thiserr[plane][col    ] += ( err * 7 ) / 16;
            fserrP->nexterr[plane][col + 2] += ( err * 3 ) / 16;
            fserrP->nexterr[plane][col + 1] += ( err * 5 ) / 16;
            fserrP->nexterr[plane][col    ] += ( err     ) / 16;
        }
    }
}



static void
floydSwitchDir(struct pam * const pamP, struct fserr * const fserrP) {

    unsigned int plane;

    for (plane = 0; plane < pamP->depth; ++plane) {
        long * const temperr = fserrP->thiserr[plane];
        fserrP->thiserr[plane] = fserrP->nexterr[plane];
        fserrP->nexterr[plane] = temperr;
    }
    fserrP->fsForward = ! fserrP->fsForward;
}



static void
searchColormap(struct pam * const pamP,
               tupletable   const colormap, 
               int          const colors,
               tuple        const tuple,
               int *        const colormapIndexP) {
/*----------------------------------------------------------------------------
   Search colormap 'colormap', which contains 'colors' colors (at
   least 1) for the color closest to that of tuple 'tuple'.  Return
   its index as *colormapIndexP.
-----------------------------------------------------------------------------*/
    unsigned int i;
    unsigned long dist;
    /* The closest distance we've found so far between the color of
       pixel *pP and a color in the colormap.  This is measured as
       the square of the cartesian distance between the colors,
       except that if maxval > 255, it's divided by 16 so that it
       will fit in a 32 bit word.  Note that each sample can be up
       to 65535 (PPM_OVERALLMAXVAL).
    */
    dist = ULONG_MAX;  /* initial value */
    for (i = 0; i < colors; ++i) {
        unsigned long newdist;  /* candidate for new 'dist' value */

        if (pamP->maxval < 256) {
            unsigned int plane;
            newdist = 0;
            for (plane=0; plane < pamP->depth; ++plane) 
                newdist += sqr(tuple[plane] - colormap[i]->tuple[plane]);
        } else {
            unsigned int plane;
            newdist = 0;
            for (plane=0; plane < pamP->depth; ++plane) 
                newdist += sqr(tuple[plane] - colormap[i]->tuple[plane]) / 16;
        }
        if (newdist < dist) {
            *colormapIndexP = i;
            dist = newdist;
        }
    }
}



static void
convertRow(struct pam * const pamP, tuple tuplerow[],
           tupletable const colormap, 
           tuplehash const colorhash, bool *usehashP,
           int const newcolors,
           bool const floyd, struct fserr * const fserrP) {
/*----------------------------------------------------------------------------
  Replace the colors in row tuplerow[] with the new colors.

  Use and update the Floyd-Steinberg state *fserrP.
-----------------------------------------------------------------------------*/
    int col;
    int limitcol;
    
    if (floyd)
        floydInitRow(pamP, fserrP);

    if ((!floyd) || fserrP->fsForward) {
        col = 0;
        limitcol = pamP->width;
    } else {
        col = pamP->width - 1;
        limitcol = -1;
    }
    do {
        int colormapIndex;
        int found;

        if (floyd) 
            floydAdjustColor(pamP, tuplerow[col], fserrP, col);

        /* Check hash table to see if we have already matched this color. */
        pnm_lookuptuple(pamP, colorhash, tuplerow[col], 
                        &found, &colormapIndex);
        if (!found)
            /* No, have to do a full lookup */
            searchColormap(pamP, colormap, newcolors, tuplerow[col], 
                           &colormapIndex);
        if (*usehashP) {
            bool fits;
            pnm_addtotuplehash(pamP, colorhash, tuplerow[col], colormapIndex, 
                               &fits);
            if (!fits) {
                pm_message("out of memory adding to hash table; "
                           "proceeding without it");
                *usehashP = FALSE;
            }
        }

        if (floyd)
            floydPropagateErr(pamP, fserrP, col, tuplerow[col], 
                              colormap[colormapIndex]->tuple);

        pnm_assigntuple(pamP, tuplerow[col], colormap[colormapIndex]->tuple);

        if (floyd && !fserrP->fsForward) 
            --col;
        else
            ++col;
    } while (col != limitcol);

    if (floyd)
        floydSwitchDir(pamP, fserrP);
}



int
main(int argc, char * argv[] ) {

    struct cmdlineInfo cmdline;
    FILE* ifp;
    struct pam inpam, outpam;
    tuple** tuples;
    int newcolors;
    tupletable colormap;
    struct fserr fserr;
    tuplehash colorhash;
    bool usehash;

    ppm_init(&argc, argv);

    initSampleCompare();
 
    parseCommandLine(argc, argv, &cmdline);

    ifp = pm_openr(cmdline.inputFilespec);

    tuples = pnm_readpam(ifp, &inpam, sizeof(inpam));
    
    pm_close(ifp);

    if (cmdline.mapFilespec) {
        FILE *mapfile;
        struct pam mappam;
        tuple ** maptuples;

        mapfile = pm_openr(cmdline.mapFilespec);
        maptuples = pnm_readpam(mapfile, &mappam, sizeof(mappam));
        pm_close(mapfile);

        adjustMaxval(&inpam, tuples, mappam.maxval); /* modifies inpam */
        adjustDepth(&inpam, &tuples, mappam.depth); /* modifies inpam */

        computeColorMapFromMap(&mappam, maptuples, &colormap, &newcolors);
        pnm_freepamarray(maptuples, &mappam);
    } else { 
        computeColorMapFromInput(&inpam, tuples, cmdline.newcolors, 
                                 cmdline.methodForLargest, 
                                 cmdline.methodForRep,
                                 &colormap, &newcolors);
            /* May modify maxval of 'tuples', inpam */
    }

    outpam = inpam; outpam.file = stdout;

    pnm_writepaminit(&outpam);

    pm_message( "mapping image to new colors..." );

    colorhash = pnm_createtuplehash();
    usehash = TRUE;

    if (cmdline.floyd)
        initFserr(&inpam, &fserr);

    {
        int row;
        for (row = 0; row < inpam.height; ++row) {
            /* The following modifies tuples[row] and both consults and 
               adds to 'colorhash' and its associated 'usehash'
            */
            convertRow(&inpam, tuples[row], colormap, colorhash, &usehash,
                       newcolors, cmdline.floyd, &fserr);
            
            pnm_writepamrow(&outpam, tuples[row]);
        }
    }
    pnm_destroytuplehash(colorhash);

    pm_close(stdout);

    exit(0);
}
